//--------------------------------------------------------------------------
//Copyright(c) 2009, WenZhong Sun<wanmyqawdr@126.com> 
//--------------------------------------------------------------------------
//This file is part of ARMV4T_and_FPGA_based_DSO
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//--------------------------------------------------------------------------
//This file contains the functions to control MMU
//--------------------------------------------------------------------------
#include "../../../inc/typedef.h"
#include "./cp15.h"

//regions definitions
static const REGION regions[] = {
    //must remap bank0 to bank6, because of the location of vectors table
    
    //do not remap bank0 when burn image to nor flash 
    {BANK0, 128, BANK0, RWRW, PTE_CB, 0},
    
    //the others keep linear map ... only can access bank6 in USER mode
    //Note: bank7 is unconnect, access bank7 will cause data abort exception.
    {BANK1, 128, BANK1, RWNA, PTE_cb, 0},
    {BANK2, 128, BANK2, RWNA, PTE_CB, 0},
    {BANK3, 128, BANK3, RWNA, PTE_CB, 0},
    {BANK4, 128, BANK4, RWNA, PTE_CB, 0},
    {BANK5, 128, BANK5, RWNA, PTE_CB, 0},
    
    //Note: access beyond address of 0x33ff_ffff will cause 
    //      data abort exception in bank 6.
    //      open write buffer in bank6 may cause lcd display errors 
    //      or some other problems,bucause the datas write to memory may be 
    //      buffered in wirte buffer, so if you use write buffer in bank6,
    //      you should clean DCache at right time.
    {BANK6, 64, BANK6, RWRW, PTE_Cb, 0},
    
    //write through, no cache in peripherals region and can not access 
    //peripherals region in USER mode, so you must access peripherals by
    //DDF, but if you realy want to do that directly, change "RWNA" to "RWRW"
    {PERIPHERAL, 3072, PERIPHERAL, RWNA, PTE_cb, 0},
    
    {0, 0, 0, 0, 0, 0}  //end
};

// *** _page_table_init ***
// -- Description  --
// fill all the PTE of L1 page table with FAULT 
// -- Parameter --
// "pt", pointer of L1 page table
// -- Return --
//      None
void _page_table_init(UINT *pt)
{
    const UINT FAULT = 0;
          UINT index = 4096;
    
    while(index--)
        pt[index] = FAULT;
}

// *** _region_init ***
// -- Description --
// fill L1 page table by regions definition
// -- Parameters --
// "pRegions", pointer of a array which contains the regions definitions
// -- Return --
//      None
void _region_init(const REGION *pRegions)
{
    UINT i;
    //first L1 page table entry
    UINT PTE_1st;
    //address of L1 page table entry
    PUINT PTEptr;
    
    while(pRegions->page_num){
        //get PTE_1st start address
        PTEptr = (PUINT)PT_BASE + (pRegions->vAddress >> 20);
        
        //clear PTE_1st 
        PTE_1st = 0;
        
        //fill first L1 page table entry
        PTE_1st |= pRegions->pAddress;  //set start physical address
        PTE_1st |= pRegions->AP << 10;  //set access permission
        PTE_1st |= pRegions->CB << 2;   //set cache and write buffer attributes
        PTE_1st |= pRegions->DOM << 5;  //set domain
        PTE_1st |= 0x12;                //set type of PTE to section 
        
        //fill others PTEs
        for(i = 0; i < pRegions->page_num; i++)
            *PTEptr++ = PTE_1st + (i << 20);
        
        //next region
        pRegions++;
    }
}

// *** _mmu_init ***
// -- Description --
// Initialize MMU before enter __main().
// -- Parameters --
//      None
// -- Return --
//      None
void _mmu_init(void)
{
    //clean and diable DCache;
    _cp15_DCache_clean();
    _cp15_DCache_disable();
    //disable MMU
    _cp15_mmu_disable();
    
    //flush and enable ICache 
    //if you disable ICache the cpu will become very slow...
    _cp15_ICache_flush();
    _cp15_ICache_enable();
    
    
    //fill all the PTE with FAULT
    _page_table_init((PUINT)PT_BASE);
    //fill L1 page table by regions definitions
    _region_init(regions);
    
    //set TTB
    _cp15_mmu_ttb_set((PUINT)PT_BASE);
    //use domain0, attribute is client
    _cp15_mmu_domain_set(CLIENT);
    //make sure don't use FCSE
    _cp15_mmu_process_id_set(0x0);   
    
    //flush TLB
    _cp15_mmu_tlb_flush();
    //enable MMU
    _cp15_mmu_enable();    

    //flush and enable DCache
    _cp15_DCache_flush();
    _cp15_DCache_enable();
}
